/*
  gsl_statistics_distributions.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <vector>
#include <cstdlib>
#include <cassert>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include "statistics.hpp"

#define numberof(a) (sizeof(a)/sizeof(a[0]))

double distribution_type_value(const size_t t, gsl_rng *r)
{
  double rv = 0;

  switch (t) {
  case 0:	rv = gsl_rng_uniform(r);	break;
  case 1:	rv = gsl_ran_ugaussian(r);	break;
  case 2:	rv = gsl_ran_gaussian(r, 1.);	break;
  case 3:	rv = gsl_ran_ugaussian_tail(r, 1.5);	break;
  case 4:	rv = gsl_ran_gaussian_tail(r, 1.5, 1.);	break;
  case 5:	rv = gsl_ran_exponential(r, 1.);	break;
  case 6:	rv = gsl_ran_laplace(r, 1.);	break;
  case 7:	rv = gsl_ran_exppow(r, 1., 2.5);	break;
  case 8:	rv = gsl_ran_cauchy(r, 1.);	break;
  case 9:	rv = gsl_ran_rayleigh(r, 1.);	break;
  case 10:	rv = gsl_ran_rayleigh_tail(r, 1., 1.);	break;
  case 11:	rv = gsl_ran_landau(r);	break;
  case 12:	rv = gsl_ran_levy(r, 1., 1.);	break;
  case 13:	rv = gsl_ran_levy_skew(r, 1., 1., 1.);	break;
  case 14:	rv = gsl_ran_gamma(r, 1., 1.);	break;
  case 15:	rv = gsl_ran_flat(r, 0., 1.);	break;
  case 16:	rv = gsl_ran_chisq(r, 1.);	break;
  case 17:	rv = gsl_ran_lognormal(r, 0., 1.);	break;
  case 18:	rv = gsl_ran_fdist(r, 3., 2.);	break;
  case 19:	rv = gsl_ran_tdist(r, 1.);	break;
  case 20:	rv = gsl_ran_beta(r, 2., 2.);	break;
  case 21:	rv = gsl_ran_logistic(r, 1.);	break;
  case 22:	rv = gsl_ran_pareto(r, 1., 1.);	break;
  case 23:	rv = gsl_ran_weibull(r, 1., 1.);	break;
  case 24:	rv = gsl_ran_gumbel1(r, 1., 1.);	break;
  case 25:	rv = gsl_ran_gumbel2(r, 1., 1.);	break;
  case 26:	rv = gsl_ran_poisson(r, 2.5);	break;
  case 27:	rv = gsl_ran_bernoulli(r, .7);	break;
  case 28:	rv = gsl_ran_binomial(r, .5, 9);	break;
  case 29:	rv = gsl_ran_negative_binomial(r, .5, 3.5);	break;
  case 30:	rv = gsl_ran_pascal(r, .5, 3);	break;
  case 31:	rv = gsl_ran_geometric(r, .5);	break;
  case 32:	rv = gsl_ran_hypergeometric(r, 5, 20, 3);	break;
  default:	rv = gsl_ran_logarithmic(r, .7);	break;
  }
  return rv;
}
const char *distribution_names[] = {
  "uniform",
  "ugaussian",
  "gaussian",
  "ugaussian_tail",
  "gaussian_tail",
  "exponential",
  "laplace",
  "exppow",
  "cauchy",
  "rayleigh",
  "rayleigh_tail",
  "landau",
  "levy",
  "levy_skew",
  "gamma",
  "flat",
  "chisq",
  "lognormal",
  "fdist",
  "tdist",
  "beta",
  "logistic",
  "pareto",
  "weibull",
  "gumbel1",
  "gumbel2",
  "poisson",
  "bernoulli",
  "binomial",
  "negative_binomial",
  "pascal",
  "geometric",
  "hypergeometric",
  "logarithmic",
};

int main()
{
  const double maximum_permissible_error(
#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
					 std::numeric_limits<double>::min()
#else
					 1e-6
#endif
					 );
  const char *result_names[] = {
    "mean",
    "variance",
    "standard deviation",
    "biased variance",
    "biased standard deviation",
    "absolute deviation",
    "skewness",
    "kurtosis",
    "lag1 autocorrelation",
    "maximum value",
    "minimum value",
  };
  const size_t n(1000000);
  double *a = (double *)malloc(sizeof(double)*n);
  const gsl_rng_type *T = gsl_rng_default;
  gsl_rng *r;

  std::string tabs_for_result_names[numberof(result_names)];
  {
    size_t l = std::string(result_names[0]).size();
    for (size_t i=1; i<numberof(result_names); ++i)
      l = std::max(l, std::string(result_names[i]).size());
    for (size_t i=0; i<numberof(result_names); ++i) {
      tabs_for_result_names[i] = "\t";
      for (size_t j=0; j<(l+1)/8 - (std::string(result_names[i]).size()+1)/8; ++j)
	tabs_for_result_names[i] += "\t";
    }
  }

  gsl_rng_env_setup();
  r = gsl_rng_alloc(T);
  for (size_t t=0; t<33; ++t) {
    std::cout << distribution_names[t] << ":" << std::endl;
    gsl_rng_set(r, gsl_rng_default_seed);
    for (size_t i=0; i<n; ++i)
      a[i] = distribution_type_value(t, r);

    double gsl_results[] = {
      gsl_stats_mean(a, 1, n),
      gsl_stats_variance(a, 1, n),
      gsl_stats_sd(a, 1, n),
      gsl_stats_variance_with_fixed_mean(a, 1, n, gsl_results[0]),
      gsl_stats_sd_with_fixed_mean(a, 1, n, gsl_results[0]),
      gsl_stats_absdev(a, 1, n),
      gsl_stats_skew(a, 1, n),
      gsl_stats_kurtosis(a, 1, n),
      gsl_stats_lag1_autocorrelation(a, 1, n),
      gsl_stats_max(a, 1, n),
      gsl_stats_min(a, 1, n),
    };

    std::vector<double> v;
    for (size_t i=0; i<n; ++i)
      v.push_back(a[i]);
    double results[] = {
      arithmetic_average(v.begin(), v.end()),
#if !0
      unbiased_variance(v.begin(), v.end(), results[0]),
      unbiased_standard_deviation(v.begin(), v.end(), results[0]),
      sample_variance(v.begin(), v.end(), results[0]),
      sample_standard_deviation(v.begin(), v.end(), results[0]),
      absolute_deviation(v.begin(), v.end(), results[0]),
      skewness(v.begin(), v.end(), results[0], results[2]),
      kurtosis(v.begin(), v.end(), results[0], results[2]),
      lag_one_autocorrelation(v.begin(), v.end(), results[0]),
#else
      unbiased_variance(v.begin(), v.end()),
      unbiased_standard_deviation(v.begin(), v.end()),
      sample_variance(v.begin(), v.end()),
      sample_standard_deviation(v.begin(), v.end()),
      absolute_deviation(v.begin(), v.end()),
      skewness(v.begin(), v.end()),
      kurtosis(v.begin(), v.end()),
      lag_one_autocorrelation(v.begin(), v.end()),
#endif
      *max_element(v.begin(), v.end()),
      *min_element(v.begin(), v.end()),
    };
    for (size_t j=0; j<numberof(results); ++j) {
      double error(sqrt(pow(gsl_results[j] - results[j], 2)));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << gsl_results[j] << "\t" << results[j] << "\t" << error << std::endl;
      assert(error < maximum_permissible_error);
    }
    std::cout << std::endl;
  }
  gsl_rng_free(r);
  return 0;
}
