/*
  statistics.hpp

  Copyright (C) 2015 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _STATISTICS_HPP_
#define _STATISTICS_HPP_
#include <functional>
#include "averages.hpp"

#if 0
#define USE_SQUARE_MEAN_REMAINDER_FUNCTION
#define USE_ABSOLUTE_MEAN_REMAINDER_FUNCTION
#define USE_CUBIC_MEAN_REMAINDER_STANDARD_DEVIATION_FUNCTION
#define USE_FOURTH_POWER_MEAN_REMAINDER_STANDARD_DEVIATION_FUNCTION
#endif

#if \
  defined(USE_SQUARE_MEAN_REMAINDER_FUNCTION) ||\
  defined(USE_ABSOLUTE_MEAN_REMAINDER_FUNCTION) ||\
  defined(USE_CUBIC_MEAN_REMAINDER_STANDARD_DEVIATION_FUNCTION) ||\
  defined(USE_FOURTH_POWER_MEAN_REMAINDER_STANDARD_DEVIATION_FUNCTION)
#define USE_IDENTITY_FUNCTION
#endif

#if defined(USE_IDENTITY_FUNCTION)
template <typename T>
struct identity_function : std::unary_function<T, T> {
  T operator()(const T &x) const
  { return x; }
};
#endif

#if defined(USE_SQUARE_MEAN_REMAINDER_FUNCTION)
template <typename T>
struct square_mean_remainder_function : std::unary_function<T, T> {
  square_mean_remainder_function(const T &mu) : mean(mu) {}
  T operator()(const T &x) const
  {
    const T v(x - mean);
    return v*v;
  }
  const T mean;
};
#endif

#if defined(USE_ABSOLUTE_MEAN_REMAINDER_FUNCTION)
template <typename T>
struct absolute_mean_remainder_function : std::unary_function<T, T> {
  absolute_mean_remainder_function(const T &mu) : mean(mu) {}
  T operator()(const T &x) const
  {
    return fabs(x - mean);
  }
  const T mean;
};
#endif

#if defined(USE_CUBIC_MEAN_REMAINDER_STANDARD_DEVIATION_FUNCTION)
template <typename T>
struct cubic_mean_remainder_standard_deviation_function : std::unary_function<T, T> {
  cubic_mean_remainder_standard_deviation_function(const T &mu, const T &sigma) : mean(mu), standard_deviation(sigma) {}
  T operator()(const T &x) const
  {
    const T v((x - mean)/standard_deviation);
    return v*v*v;
  }
  const T mean, standard_deviation;
};
#endif

#if defined(USE_FOURTH_POWER_MEAN_REMAINDER_STANDARD_DEVIATION_FUNCTION)
template <typename T>
struct fourth_power_mean_remainder_standard_deviation_function : std::unary_function<T, T> {
  fourth_power_mean_remainder_standard_deviation_function(const T &mu, const T &sigma) : mean(mu), standard_deviation(sigma) {}
  T operator()(const T &x) const
  {
    const T v((x - mean)/standard_deviation);
    return v*v*v*v;
  }
  const T mean, standard_deviation;
};
#endif

template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type sample_variance(const InputIterator first, const InputIterator last, const typename std::iterator_traits<InputIterator>::value_type mean)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
#if defined(USE_SQUARE_MEAN_REMAINDER_FUNCTION)
  return generalized_average(first, last, square_mean_remainder_function<value_type>(mean), identity_function<value_type>());
#else
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
  while (it != last) {
    INTERNAL_VALUE_TYPE_OF_ITERATOR delta(*it - mean);
    v += (delta*delta - v)/std::distance(first, ++it);
  }
  return v;
#endif
}
template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type sample_variance(const InputIterator first, const InputIterator last)
{
  return sample_variance(first, last, arithmetic_average(first, last));
}

template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type unbiased_variance(const InputIterator first, const InputIterator last, const typename std::iterator_traits<InputIterator>::value_type mean)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
  return sample_variance(first, last, mean)*(value_type(std::distance(first, last))/(std::distance(first, last) - 1));
}
template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type unbiased_variance(const InputIterator first, const InputIterator last)
{
  return unbiased_variance(first, last, arithmetic_average(first, last));
}

template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type sample_standard_deviation(const InputIterator first, const InputIterator last, const typename std::iterator_traits<InputIterator>::value_type mean)
{
  return sqrt(sample_variance(first, last, mean));
}
template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type sample_standard_deviation(const InputIterator first, const InputIterator last)
{
  return sample_standard_deviation(first, last, arithmetic_average(first, last));
}

template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type unbiased_standard_deviation(const InputIterator first, const InputIterator last, const typename std::iterator_traits<InputIterator>::value_type mean)
{
  return sqrt(unbiased_variance(first, last, mean));
}
template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type unbiased_standard_deviation(const InputIterator first, const InputIterator last)
{
  return unbiased_standard_deviation(first, last, arithmetic_average(first, last));
}

template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type absolute_deviation(const InputIterator first, const InputIterator last, const typename std::iterator_traits<InputIterator>::value_type mean)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
#if defined(USE_ABSOLUTE_MEAN_REMAINDER_FUNCTION)
  return generalized_average(first, last, absolute_mean_remainder_function<value_type>(mean), identity_function<value_type>());
#else
  InputIterator it(first);
  value_type v(0);
  while (it != last)
    v += fabs(*it++ - mean);
  return v/std::distance(first, last);
#endif
}
template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type absolute_deviation(const InputIterator first, const InputIterator last)
{
  return absolute_deviation(first, last, arithmetic_average(first, last));
}

template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type skewness(const InputIterator first, const InputIterator last, const typename std::iterator_traits<InputIterator>::value_type mean, const typename std::iterator_traits<InputIterator>::value_type standard_deviation)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
#if defined(USE_CUBIC_MEAN_REMAINDER_STANDARD_DEVIATION_FUNCTION)
  return generalized_average(first, last, cubic_mean_remainder_standard_deviation_function<value_type>(mean, standard_deviation), identity_function<value_type>());
#else
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
  while (it != last) {
    INTERNAL_VALUE_TYPE_OF_ITERATOR w((*it - mean)/standard_deviation);
    v += (w*w*w - v)/std::distance(first, ++it);
  }
  return v;
#endif
}
template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type skewness(const InputIterator first, const InputIterator last)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
  const value_type mean(arithmetic_average(first, last));
  return skewness(first, last, mean, unbiased_standard_deviation(first, last, mean));
}

template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type third_moment(const InputIterator first, const InputIterator last, const typename std::iterator_traits<InputIterator>::value_type mean)
{
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
  while (it != last) {
    INTERNAL_VALUE_TYPE_OF_ITERATOR w(*it - mean);
    v += (w*w*w - v)/std::distance(first, ++it);
  }
  return v;
}
template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type third_moment(const InputIterator first, const InputIterator last)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
  const value_type mean(arithmetic_average(first, last));
  return third_moment(first, last, mean);
}

template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type kurtosis(const InputIterator first, const InputIterator last, const typename std::iterator_traits<InputIterator>::value_type mean, const typename std::iterator_traits<InputIterator>::value_type standard_deviation)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
#if defined(USE_FOURTH_POWER_MEAN_REMAINDER_STANDARD_DEVIATION_FUNCTION)
  return generalized_average(first, last, fourth_power_mean_remainder_standard_deviation_function<value_type>(mean, standard_deviation), identity_function<value_type>()) - 3;
#else
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
  while (it != last) {
    INTERNAL_VALUE_TYPE_OF_ITERATOR w((*it - mean)/standard_deviation);
    v += (w*w*w*w - v)/std::distance(first, ++it);
  }
  return v - 3;
#endif
}
template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type kurtosis(const InputIterator first, const InputIterator last)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
  const value_type mean(arithmetic_average(first, last));
  return kurtosis(first, last, mean, unbiased_standard_deviation(first, last, mean));
}

template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type fourth_moment(const InputIterator first, const InputIterator last, const typename std::iterator_traits<InputIterator>::value_type mean)
{
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
  while (it != last) {
    INTERNAL_VALUE_TYPE_OF_ITERATOR w((*it - mean));
    v += (w*w*w*w - v)/std::distance(first, ++it);
  }
  return v;
}
template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type fourth_moment(const InputIterator first, const InputIterator last)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
  const value_type mean(arithmetic_average(first, last));
  return fourth_moment(first, last, mean);
}

template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type lag_one_autocorrelation(const InputIterator first, const InputIterator last, const typename std::iterator_traits<InputIterator>::value_type mean)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
  InputIterator it(first);
  value_type y(*it - mean);
  INTERNAL_VALUE_TYPE_OF_ITERATOR n(0), d(y*y);
  while (it+1 != last) {
    INTERNAL_VALUE_TYPE_OF_ITERATOR x(*it++ - mean), y(*it - mean);
#if 0
    n += x*y;
    d += y*y;
#else
    n += (x*y - n)/(std::distance(first, it) + 1);
    d += (y*y - d)/(std::distance(first, it) + 1);
#endif
  }
  return n/d;
}
template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type lag_one_autocorrelation(const InputIterator first, const InputIterator last)
{
  return lag_one_autocorrelation(first, last, arithmetic_average(first, last));
}

#endif
